
/* tcptrunc
 * Mark Allman (mallman@icir.org)
 * 
 * Copyright (c) 2016 International Computer Science Institute
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation files
 * (the "Software"), to deal in the Software without restriction,
 * including without limitation the rights to use, copy, modify, merge,
 * publish, distribute, sublicense, and/or sell copies of the Software,
 * and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * The names and trademarks of copyright holders may not be used in
 * advertising or publicity pertaining to the software without specific
 * prior permission. Title to copyright in this software and any
 * associated documentation will at all times remain with the copyright
 * holders.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <pcap.h>

#define FALSE                  0
#define TRUE                   1
#define EH_SIZE                14
#define EH_TYPE_OFFSET         12

char *infile = "-";
char *outfile = "-";
pcap_t *inf = NULL;
pcap_dumper_t *outf = NULL;
unsigned char skip_unknown = FALSE;

void usage (char *progname)
{
    fprintf (stderr,"usage: tcptrunc [options]\n");
    fprintf (stderr,"       -h        usage instructions\n");
    fprintf (stderr,"       -r file   input packet trace (default: stdin)\n");
    fprintf (stderr,"       -s        skip non-UDP and non-TCP packets\n");
    fprintf (stderr,"       -w file   output packet trace (default: stdout)\n");
    fprintf (stderr,"       -v        version information\n");
    exit (1);
}

void dump_version ()
{
    fprintf (stdout,"tcptrunc v%s\n", VERSION);
    exit (0);
}
        
void parseargs (int argc, char *argv [])
{
    int opt;

    while ((opt = getopt (argc, argv, "r:w:hsv")) != -1)
    {
        switch (opt)
        {
            case 'h':
              usage (argv [0]);
            case 'r':
              infile = optarg;
              break;
            case 's':
              skip_unknown = TRUE;
              break;
            case 'v':
              dump_version ();
            case 'w':
              outfile = optarg;
              break;
            default:
              usage (argv [0]);
        }
    }
}

void open_trace_files ()
{
    char errbuf [PCAP_ERRBUF_SIZE];

    if ((inf = pcap_open_offline (infile, errbuf)) == NULL)
    {
	fprintf (stderr,"error opening tracefile %s: %s\n", infile, errbuf);
	exit (1);
    }
    if ((outf = pcap_dump_open (inf,outfile)) == NULL)
    {
        fprintf (stderr,"cannot open %s for writing\n", outfile);
        exit (1);
    }
}

void dump_pkt (unsigned char unknown,struct pcap_pkthdr *hdr,u_char *pkt)
{
    if (unknown && skip_unknown)
        return;
    pcap_dump ((u_char *)outf,hdr,pkt);
}

void strip_payload (unsigned short offset,struct pcap_pkthdr *hdr,u_char *pkt)
{
    memset (pkt + offset,'\0',hdr->caplen - offset);
    hdr->caplen = offset;
}

void process_trace ()
{
    struct pcap_pkthdr hdr;
    struct ip *iph;
    struct tcphdr *tcph;
    u_char *pkt;
    unsigned short offset;
    unsigned short found_ip_hdr;
    unsigned short proc_etherhdr;
    unsigned short eth_type;
    unsigned short tcp_hl;

    while ((pkt = (u_char *)pcap_next (inf,&hdr)) != NULL)
    {
        if (hdr.caplen < EH_SIZE)
        {
            dump_pkt (TRUE,&hdr,(u_char *)pkt);
            continue;
        }
        /* VLAN tag processing from Robert Bullen */
        offset = EH_TYPE_OFFSET;
        proc_etherhdr = TRUE;
        found_ip_hdr = FALSE;
        while (proc_etherhdr)
        {
            if (hdr.caplen < (offset + sizeof (unsigned short)))
                break;
            eth_type = ntohs (*((unsigned short*) (pkt + offset)));
            switch (eth_type)
            {
                case 0x0800:      /* IPv4 */
                    proc_etherhdr = FALSE;
                    found_ip_hdr = TRUE;
                    offset += 2;
                    break;
                case 0x8100:      /* IEEE 802.1Q (VLAN-tagging TPID) */
                case 0x88a8:      /* IEEE 802.1ad (provider-tagging TPID) */
                case 0x9100:      /* IEEE 802.1QinQ (double-tagging TPID) */
                case 0x9200:      /* IEEE 802.1QinQ (alternate TPID) */
                case 0x9300:      /* IEEE 802.1QinQ (alternate TPID) */
                    offset += 4;
                    break;
                case 0x8926:      /* Cisco VNTag (for FET and UMC). */
                    offset += 6;
                    break;
                default:
                    proc_etherhdr = FALSE;
                    found_ip_hdr = FALSE;
                    break;
            }
        }
        if (!found_ip_hdr || (hdr.caplen < (offset + sizeof (struct ip))))
        {
            dump_pkt (TRUE,&hdr,(u_char *)pkt);
            continue;
        }
        iph = (struct ip *)(pkt + offset);
        if ((ntohs (iph->ip_off) & IP_OFFMASK) != 0)
        {
            /* if this is a fragment, but not the first fragment, so dump 
               only the IP header */ 
            offset += (iph->ip_hl * 4);
            if (hdr.caplen > offset)
                strip_payload (offset,&hdr,pkt);
            dump_pkt (((iph->ip_p == IPPROTO_TCP) ||
                       (iph->ip_p == IPPROTO_UDP)) ? FALSE : TRUE,
                      &hdr,(u_char *)pkt);
            continue;
        }
        switch (iph->ip_p)
        {
            case IPPROTO_TCP:
              offset += (iph->ip_hl * 4);
              tcph = (struct tcphdr *)(pkt + offset);
              if (hdr.caplen > (offset + sizeof (struct tcphdr)))
              {
                  tcp_hl = 4 * tcph->th_off;
                  offset += (tcp_hl > 20 ? tcp_hl : 20);
                  if (hdr.caplen > offset)
                      strip_payload (offset,&hdr,pkt);
              }
              dump_pkt (FALSE,&hdr,(u_char *)pkt);
              break;
            case IPPROTO_UDP:
              offset += ((iph->ip_hl * 4) + sizeof (struct udphdr));
              if (hdr.caplen > offset)
                  strip_payload (offset,&hdr,pkt);
              dump_pkt (FALSE,&hdr,(u_char *)pkt);
              break;
            default:
              dump_pkt (TRUE,&hdr,(u_char *)pkt);
        }
    }
}

void close_trace_files ()
{
    pcap_close (inf);
    pcap_dump_close (outf);
}

int main (int argc,char *argv [])
{
    parseargs (argc,argv);
    open_trace_files ();
    process_trace ();
    close_trace_files ();
    exit (0);
}
